from __future__ import print_function
from types import ModuleType
import logging

import rdflib
from rdflib.term import Variable, URIRef
from rdflib.graph import ConjunctiveGraph
import wrapt

from .data import DataUser

from .context_common import CONTEXT_IMPORTS
from .context_store import ContextStore, RDFContextStore
from .contextualize import (BaseContextualizable,
                            Contextualizable,
                            ContextualizableClass,
                            ContextualizingProxy,
                            contextualize_metaclass)
from .utils import FCN
from six.moves.urllib.parse import quote
from six import text_type
import six


L = logging.getLogger(__name__)

DEFAULT_CONTEXT_KEY = 'default_context_id'
'''
Configuration file key for the URI of a default RDF graph context.

This is the URI of the default graph in a project or bundle.
'''

IMPORTS_CONTEXT_KEY = 'imports_context_id'
'''
Configuration file key for the URI of an imports RDF graph context.

The imports context holds the relationships between contexts, especially the imports
relationship
'''

# TODO: Move this into mapper or a new mapper_common module
CLASS_REGISTRY_CONTEXT_KEY = 'class_registry_context_id'
'''
Configuration file key for the URI of the class registry RDF graph context.

The class registry context holds the mappings between RDF types and Python classes for a
project or bundle.
'''


class ModuleProxy(wrapt.ObjectProxy):
    def __init__(self, ctx, *args, **kwargs):
        super(ModuleProxy, self).__init__(*args, **kwargs)
        self._self_overrides = dict()
        self._self_ctx = ctx

    def add_attr_override(self, name, override):
        self._self_overrides[name] = override

    def __getattr__(self, name):
        o = self._self_overrides.get(name, None)
        if o is not None:
            return o
        else:
            o = super(ModuleProxy, self).__getattr__(name)
            if isinstance(o, (BaseContextualizable, ContextualizableClass)):
                o = o.contextualize(self._self_ctx)
                self._self_overrides[name] = o
            return o


class ContextMeta(ContextualizableClass):
    @property
    def context(self):
        return None

    @context.setter
    def context(self, v):
        pass

    def contextualize_class_augment(self, context):
        if context is None:
            return self
        ctxd_meta = contextualize_metaclass(context, self)
        res = ctxd_meta(self.__name__, (self,), dict(class_context=context.identifier))
        res.__module__ = self.__module__
        return res


class ContextualizableDataUserMixin(Contextualizable, DataUser):

    @property
    def conf(self):
        if self.context is None:
            return super(ContextualizableDataUserMixin, self).conf
        else:
            return self.context.conf

    @conf.setter
    def conf(self, conf):
        super(ContextualizableDataUserMixin, self).conf = conf


class Context(six.with_metaclass(ContextMeta,
                                 ContextualizableDataUserMixin)):
    """
    A context. Analogous to an RDF context, with some special sauce

    .. automethod:: __call__
    .. automethod:: __bool__
    """

    def __init__(self, ident=None,
                 imported=(),
                 mapper=None,
                 key=None,
                 base_namespace=None,
                 **kwargs):
        super(Context, self).__init__(**kwargs)

        if key is not None and ident is not None:
            raise Exception("Only one of 'key' or 'ident' can be given to Context")
        if key is not None and base_namespace is None:
            raise Exception("If 'key' is given, then 'base_namespace' must also be given to Context")

        if not isinstance(ident, URIRef) \
           and isinstance(ident, (str, text_type)):
            ident = URIRef(ident)

        if not isinstance(base_namespace, rdflib.namespace.Namespace) \
           and isinstance(base_namespace, (str, text_type)):
            base_namespace = rdflib.namespace.Namespace(base_namespace)

        if ident is None and key is not None:
            ident = URIRef(base_namespace[quote(key)])

        if not hasattr(self, 'identifier'):
            self.identifier = ident
        else:
            raise Exception(self)

        self._statements = []
        self._imported_contexts = list(imported)
        self._rdf_object = None
        self._graph = None

        if mapper is None:
            my_context = getattr(self, 'context')
            if my_context:
                mapper = my_context.mapper

        self.__mapper = mapper
        self.base_namespace = base_namespace

        self._change_counter = 0
        self._triples_saved = 0

        self._stored_context = None
        self._own_stored_context = None

    @property
    def mapper(self):
        if self.__mapper is None:
            from .mapper import Mapper
            self.__mapper = Mapper(conf=self.conf)

        return self.__mapper

    def contents(self):
        '''
        Returns statements added to this context

        Returns
        -------
        generator
        '''
        return (x for x in self._statements)

    def clear(self):
        '''
        Clear declared statements
        '''
        del self._statements[:]

    def add_import(self, context):
        '''
        Add an imported context
        '''
        self._imported_contexts.append(context)

    def add_statement(self, stmt):
        '''
        Add a statement to the context. Typically, statements will be added by
        `contextualizing <Contextualizable.contextualize>` a
        `~owmeta_core.dataobject.DataObject` and making a statement thereon. For instance,
        if a class ``A`` has a property ``p``, then for the context ``ctx``::

            ctx(A)(ident='http://example.org').p('val')

        would add a statement to ``ctx`` like::

            (A(ident='http://example.org'), A.p.link, rdflib.term.Literal('val'))

        Parameters
        ----------
        stmt : tuple
            Statement to add
        '''
        if self.identifier != stmt.context.identifier:
            raise ValueError("Cannot add statements from a different context")
        self._graph = None
        self._statements.append(stmt)
        self._change_counter += 1

    def remove_statement(self, stmt):
        '''
        Remove a statement from the context

        Parameters
        ----------
        stmt : tuple
            Statement to remove
        '''
        self._graph = None
        self._statements.remove(stmt)
        self._change_counter += 1

    @property
    def imports(self):
        '''
        Return imports on this context

        Yields
        ------
        Context
        '''
        for x in self._imported_contexts:
            yield x

    def transitive_imports(self):
        '''
        Return imports on this context and on imported contexts

        Yields
        ------
        Context
        '''
        for x in self._imported_contexts:
            yield x
            for y in x.transitive_imports():
                yield y

    def save_imports(self, context=None, *args, transitive=True, **kwargs):
        '''
        Add the `imports` on this context to a graph

        Parameters
        ----------
        context : .Context, optional
            The context to add statements to. This context's configured graph will
            ultimately receive the triples. By default, a context will be created with
            ``self.conf[IMPORTS_CONTEXT_KEY]`` as the identifier
        transitive : bool, optional
            If `True`, call imported imported contexts to save their imports as well
        '''
        if not context:
            ctx_key = self.conf[IMPORTS_CONTEXT_KEY]
            context = Context(ident=ctx_key, conf=self.conf)
        self.declare_imports(context, transitive)
        context.save_context(*args, **kwargs)

    def declare_imports(self, context=None, transitive=False):
        '''
        Declare `imports <~context_dataobject.ContextDataObject.imports>` statements in
        the given context

        Parameters
        ----------
        context : .Context, optional
            The context in which to declare statements. If not provided, one will be
            created with ``self.conf[IMPORTS_CONTEXT_KEY]`` as the identifier

        Returns
        -------
        Context
            The context in which the statements were declared
        '''
        if not context:
            ctx_key = self.conf[IMPORTS_CONTEXT_KEY]
            context = Context(ident=ctx_key, conf=self.conf)
        self._declare_imports(context, transitive)
        return context

    def _declare_imports(self, context, transitive):
        for ctx in self._imported_contexts:
            if self.identifier is not None \
                    and ctx.identifier is not None \
                    and not isinstance(ctx.identifier, rdflib.term.BNode):
                context(self.rdf_object).imports(ctx.rdf_object)
                if transitive:
                    ctx._declare_imports(context, transitive)

    def save_context(self, graph=None, inline_imports=False, autocommit=True, saved_contexts=None):
        '''
        Adds the staged statements in the context to a graph

        Parameters
        ----------
        graph : rdflib.graph.Graph or set, optional
            the destination graph. Defaults to ``self.rdf``
        inline_imports : bool, optional
            if `True`, imported contexts will also be written added to the graph
        autocommit : boolean, optional
            if `True`, `graph.commit <rdflib.graph.Graph.commit>` is invoked after adding statements to the
            graph (including any imported contexts if `inline_imports` is `True`)
        saved_contexts : set, optional
            a collection of identifiers for previously saved contexts. Note that `id` is
            used to get an identifier: the return value of `id` can be repeated after an
            object is deleted.
        '''
        if saved_contexts is None:
            saved_contexts = set()

        if (self._change_counter, id(self)) in saved_contexts:
            return

        saved_contexts.add((self._change_counter, id(self)))

        if graph is None:
            graph = self._retrieve_configured_graph()
        if autocommit and hasattr(graph, 'commit'):
            graph.commit()

        if inline_imports:
            for ctx in self._imported_contexts:
                ctx.save_context(graph, inline_imports, autocommit=False,
                        saved_contexts=saved_contexts)

        # XXX: Why is this here....?
        if hasattr(graph, 'bind') and self.mapper is not None:
            for c in self.mapper.mapped_classes():
                if hasattr(c, 'rdf_namespace'):
                    try:
                        graph.bind(c.__name__, c.rdf_namespace)
                    except Exception:
                        L.warning('Failed to bind RDF namespace for %s to %s', c.__name__,
                               c.rdf_namespace, exc_info=True)
        if isinstance(graph, set):
            graph.update(self._save_context_triples())
        else:
            ctx_graph = self.get_target_graph(graph)
            ctx_graph.addN((s, p, o, ctx_graph) for s, p, o in self._save_context_triples())

        if autocommit and hasattr(graph, 'commit'):
            graph.commit()

    save = save_context
    ''' Alias to save_context '''

    @property
    def triples_saved(self):
        '''
        The number of triples saved in the most recent call to `save_context`
        '''
        return self._triples_saved_helper()

    def _triples_saved_helper(self, seen=None):
        if seen is None:
            seen = set()
        if id(self) in seen:
            return 0
        seen.add(id(self))
        res = self._triples_saved
        for ctx in self._imported_contexts:
            res += ctx._triples_saved_helper(seen)
        return res

    def _save_context_triples(self):
        self._triples_saved = 0
        for x in self._statements:
            t = x.to_triple()
            if not (isinstance(t[0], Variable) or
                    isinstance(t[2], Variable) or
                    isinstance(t[1], Variable)):
                self._triples_saved += 1
                yield t

    def get_target_graph(self, graph):
        res = graph
        if self.identifier is not None:
            if hasattr(graph, 'graph_aware') and graph.graph_aware:
                res = graph.graph(self.identifier)
            elif hasattr(graph, 'context_aware') and graph.context_aware:
                res = graph.get_context(self.identifier)
        return res

    def contents_triples(self):
        '''
        Returns, as triples, the statements staged in this context

        Yields
        ------
        tuple
            A triple of `RDFLib Identifiers <rdflib.term.Identifier>`

        '''
        for x in self._statements:
            yield x.to_triple()

    def contextualize_augment(self, context):
        '''
        Returns a contextualized proxy of this context

        Parameters
        ----------
        context : .Context
            The context to contextualize this context with
        '''
        res = ContextualizingProxy(context, self)
        res.add_attr_override('_stored_context', None)
        res.add_attr_override('_own_stored_context', None)
        return res

    @property
    def rdf_object(self):
        '''
        Returns a dataobject for this context

        Returns
        -------
        owmeta_core.dataobject.DataObject
        '''
        if self._rdf_object is None:
            from owmeta_core.context_dataobject import ContextDataObject
            self._rdf_object = ContextDataObject.contextualize(self.context)(ident=self.identifier)

        return self._rdf_object.contextualize(self.context)

    def __bool__(self):
        '''
        Always returns `True`. Prevents a context with zero statements from testing false
        since that's not typically a useful branching condition.
        '''
        return True

    __nonzero__ = __bool__

    def __len__(self):
        return len(self._statements)

    def __call__(self, o=None, *args, **kwargs):
        """
        Contextualize an object

        Parameters
        ----------
        o : object
            The object to contexualize
        """
        if o is None:
            if kwargs:
                o = kwargs
        elif args:
            o = {x.__name__: x for x in [o] + list(args)}

        if isinstance(o, ModuleType):
            return ModuleProxy(self, o)
        elif isinstance(o, dict):
            return ContextContextManager(self, o)
        elif isinstance(o, BaseContextualizable):
            return o.contextualize(self)
        elif isinstance(o, ContextualizableClass):
            # Yes, you can call contextualize on a class and it'll do the right
            # thing, but let's keep it simple here, okay?
            return o.contextualize_class(self)
        else:
            return o

    def __str__(self):
        return repr(self)

    def __repr__(self):
        ident = getattr(self, 'identifier', '???')
        if ident is None:
            identpart = ''
        else:
            identpart = 'ident="{}"'.format(ident)
        return '{}({})'.format(FCN(type(self)), identpart)

    def load_own_graph_from_configured_store(self):
        '''
        Create a RDFLib graph for accessing statements in this context, *excluding* imported
        contexts. The "configured" graph is the one at ``self.conf['rdf.graph']``.

        Returns
        -------
        rdflib.graph.ConjunctiveGraph
        '''
        return ConjunctiveGraph(identifier=self.identifier,
                                store=RDFContextStore(self, include_imports=False))

    def load_graph_from_configured_store(self):
        '''
        Create an RDFLib graph for accessing statements in this context, *including* imported
        contexts. The "configured" graph is the one at ``self.rdf``.

        Returns
        -------
        rdflib.graph.ConjunctiveGraph
        '''
        return ConjunctiveGraph(identifier=self.identifier,
                store=RDFContextStore(self, imports_graph=self._imports_graph()))

    def _imports_graph(self):
        ctxid = self.conf.get(IMPORTS_CONTEXT_KEY, None)
        return ctxid and self.rdf.get_context(URIRef(ctxid))

    def rdf_graph(self):
        '''
        Return the principal graph for this context. For a regular `Context` this will be
        the "staged" graph.

        Returns
        -------
        rdflib.graph.ConjunctiveGraph

        See Also
        --------
        staged : Has the "staged" principal graph.
        mixed : Has the "mixed" principal graph.
        stored : Has the "stored" graph, including imports.
        own_stored : Has the "stored" graph, excluding imports.
        '''
        if self._graph is None:
            self._graph = self.load_staged_graph()
        return self._graph

    def load_mixed_graph(self):
        '''
        Create a graph for accessing statements both staged (see `load_staged_graph`) and
        stored (see `load_graph_from_configured_store`). No effort is made to either
        deduplicate, smush blank nodes, or logically reconcile statements between staged
        and stored graphs.

        Returns
        -------
        rdflib.graph.ConjunctiveGraph
        '''
        return ConjunctiveGraph(identifier=self.identifier,
                                store=ContextStore(context=self, include_stored=True,
                                    imports_graph=self._imports_graph()))

    def load_staged_graph(self):
        '''
        Create a graph for accessing statements declared in this specific instance of
        this context. This statements may not have been written to disk; therefore, they
        are "staged".

        Returns
        -------
        rdflib.graph.ConjunctiveGraph
        '''
        return ConjunctiveGraph(identifier=self.identifier, store=ContextStore(context=self))

    @property
    def mixed(self):
        '''
        A read-only context whose principal graph is the "mixed" graph.

        Returns
        -------
        QueryContext

        See also
        --------
        rdf_graph
        load_mixed_graph : Defines the principal graph for this context
        '''
        return QueryContext(
                mapper=self.mapper,
                graph=self.load_mixed_graph(),
                ident=self.identifier,
                conf=self.conf)

    @property
    def staged(self):
        '''
        A read-only context whose principal graph is the "staged" graph.

        Returns
        -------
        QueryContext

        See also
        --------
        rdf_graph
        load_staged_graph : Defines the principal graph for this context
        '''
        return QueryContext(
                mapper=self.mapper,
                graph=self.load_staged_graph(),
                ident=self.identifier,
                conf=self.conf)

    @property
    def stored(self):
        '''
        A read-only context whose principal graph is the "stored" graph, including
        imported contexts.

        Returns
        -------
        QueryContext

        See also
        --------
        rdf_graph
        load_graph_from_configured_store : Defines the principal graph for this context
        '''
        if self._stored_context is None:
            self._stored_context = QueryContext(
                    mapper=self.mapper,
                    graph=self.load_graph_from_configured_store(),
                    ident=self.identifier,
                    conf=self.conf)
        return self._stored_context

    @property
    def own_stored(self):
        '''
        A read-only context whose principal graph is the "stored" graph, excluding
        imported contexts.

        Returns
        -------
        QueryContext

        See also
        --------
        rdf_graph
        load_own_graph_from_configured_store : Defines the principal graph for this context
        '''
        if self._own_stored_context is None:
            self._own_stored_context = QueryContext(
                    mapper=self.mapper,
                    graph=self.load_own_graph_from_configured_store(),
                    ident=self.identifier,
                    conf=self.conf)
        return self._own_stored_context

    def _retrieve_configured_graph(self):
        return self.rdf

    def resolve_class(self, uri):
        if self.mapper is None:
            return None
        return self.mapper.resolve_class(uri, self)


class QueryContext(Context):
    '''
    A read-only context.
    '''
    def __init__(self, graph, *args, **kwargs):
        super(QueryContext, self).__init__(*args, **kwargs)
        self.__graph = graph

    def rdf_graph(self, *args, **kwargs):
        return self.__graph

    @property
    def imports(self):
        ctxid = self.conf.get(IMPORTS_CONTEXT_KEY, None)
        imports_graph = ctxid and self.__graph.get_context(URIRef(ctxid))
        if imports_graph is None:
            return
        for t in imports_graph.triples((self.identifier, CONTEXT_IMPORTS, None)):
            yield QueryContext(mapper=self.mapper,
                    graph=self.__graph,
                    ident=t[2],
                    conf=self.conf)

    def add_import(self, *args, **kwargs): raise ContextIsReadOnly
    def save_imports(self, *args, **kwargs): raise ContextIsReadOnly
    def save_context(self, *args, **kwargs): raise ContextIsReadOnly


class ContextIsReadOnly(Exception):
    def __init__(self):
        super(ContextIsReadOnly, self).__init__('This context is read-only')


ClassContexts = dict()


class ClassContextMeta(ContextMeta):

    def __call__(self, ident, base_namespace=None, imported=()):
        res = ClassContexts.get(URIRef(ident))
        if not res:
            res = super(ClassContextMeta, self).__call__(ident=ident,
                    base_namespace=base_namespace, imported=imported)
            ClassContexts[URIRef(ident)] = res
        else:
            if base_namespace or imported:
                raise Exception('Arguments can only be provided to a ClassContext on'
                                ' first creation')
        return res


class ClassContext(six.with_metaclass(ClassContextMeta, Context)):
    pass


class ContextContextManager(object):
    """ The context manager created when Context::__call__ is passed a dict """

    def __init__(self, ctx, to_import):
        self._overrides = dict()
        self._ctx = ctx
        self._backing_dict = to_import
        self.save = self._ctx.save_context

    @property
    def context(self):
        return self._ctx

    def __call__(self, o):
        return self._ctx(o)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __getattr__(self, name):
        return self.lookup(name)

    def __getitem__(self, key):
        return self.lookup(key)

    def lookup(self, key):
        o = self._overrides.get(key, None)
        if o is not None:
            return o
        o = self._backing_dict[key]
        if isinstance(o, (BaseContextualizable, ContextualizableClass)):
            o = o.contextualize(self._ctx)
            self._overrides[key] = o
        return o



'''
Data translators

Some :class:`~owmeta_core.datasource.DataSource` and `~owmeta_core.datasource.DataTranslator`
types. Some deal with generic file types (e.g., comma-separated values) while others are
specific to the format of a kind of file housed in owmeta.
'''



from json import dumps
from weakref import WeakValueDictionary
import logging

from rdflib import plugin
from rdflib.store import Store, VALID_STORE

from .utils import FCN


L = logging.getLogger(__name__)

RDFLIB_PLUGIN_KEY = 'owmeta_core_bds'


class BundleDependencyStore(Store):
    '''
    A read-only RDFLib `~rdflib.store.Store` that supports the extra stuff we need from
    dependencies
    '''

    context_aware = True
    '''
    Specified by RDFLib. Required to be True for `~rdflib.graph.ConjunctiveGraph` stores.

    Wrapped store MUST be context-aware. This is enforced by :meth:`open`.
    '''

    def __init__(self, wrapped=None, excludes=()):
        self.wrapped = wrapped
        self.excludes = set(excludes)
        self._store_cache = None
        self._store_cache_key = None
        self._store_cache_wrapped_key = None

    def open(self, configuration):
        '''
        Creates and opens the configured store.

        Also verifies that the provided store is context-aware
        '''
        store_cache = None
        if isinstance(configuration, (list, tuple)):
            try:
                store_key, store_conf = configuration
            except ValueError:
                raise ValueError('Configuration should be in the form (store_type,'
                        ' store_configuration)')
        elif isinstance(configuration, dict):
            try:
                store_key = configuration['type']
                store_conf = configuration['conf']
            except KeyError:
                raise ValueError('Missing type and conf entries')
            self.excludes = configuration.get('excludes', ())
            store_cache = configuration.get('cache', None)
            self._store_cache = store_cache
        else:
            raise ValueError('Invalid configuration for ' + RDFLIB_PLUGIN_KEY)

        # Check for a cached BDS -- we can use that as our wrapped store, possibly
        # ourselves if an attempt is made to re-open this store.
        if store_cache and _is_cacheable(RDFLIB_PLUGIN_KEY, configuration):
            bds_ck = _cache_key(RDFLIB_PLUGIN_KEY, configuration)
            self._store_cache_key = bds_ck
            bds_cached_store = store_cache.get(bds_ck, None)
            if bds_cached_store is not None:
                store_cache.check_out(bds_ck)
                if bds_cached_store is not self:
                    self.wrapped = bds_cached_store
                    self._store_cache_wrapped_key = bds_ck
                    store_cache.check_out(bds_ck)
                # We've already opened the primary store for this config and put it in the
                # cache, so nothing left to do...
                return VALID_STORE

        if store_cache and _is_cacheable(store_key, store_conf):
            ck = _cache_key(store_key, store_conf)
            self._store_cache_wrapped_key = ck
            cached_store = store_cache.get(ck, None)
            if cached_store is None:
                cached_store = plugin.get(store_key, Store)()
                cached_store.open(store_conf)
                store_cache[ck] = cached_store
            store_cache.check_out(ck)
            self.wrapped = cached_store
        else:
            self.wrapped = plugin.get(store_key, Store)()
            self.wrapped.open(store_conf)

        assert self.wrapped.context_aware, 'Wrapped store must be context-aware.'
        self.supports_range_queries = getattr(self.wrapped, 'supports_range_queries',
                False)
        if store_cache and _is_cacheable(RDFLIB_PLUGIN_KEY, configuration):
            # If there were a stored cacheable configuration with the same cache key as
            # ours, we would have already set it as our wrapped (or found we are the
            # primary for that configuration) and returned, so we can safely set ourselves
            # in the cache
            store_cache[bds_ck] = self
            store_cache.check_out(bds_ck)

        return VALID_STORE

    def close(self, commit_pending_transaction=False):
        if self._store_cache is not None:
            if self._store_cache_key is not None:
                refcount = self._store_cache.check_in(self._store_cache_key)
                if refcount == 0:
                    if self._store_cache_wrapped_key is None:
                        msg = ('A wrapped store key is not available for a cacheable'
                               ' BDS: this should never happen')
                        L.error(msg)
                        raise Exception(msg)
                    self._close_cached_wrapped(commit_pending_transaction=commit_pending_transaction)
                else:
                    L.debug("BDS store is still referenced %d times in the cache as %s. Cannot close it yet",
                            refcount, self._store_cache_key)
            elif self._store_cache_wrapped_key is not None:
                self._close_cached_wrapped(commit_pending_transaction=commit_pending_transaction)
            else:
                # We didn't get the wrapped store from the cache/put it into a cache, so
                # we can just close it
                self.wrapped.close(commit_pending_transaction=commit_pending_transaction)
        else:
            L.debug("Closing wrapped store %s", self.wrapped)
            self.wrapped.close(commit_pending_transaction=commit_pending_transaction)

    def _close_cached_wrapped(self, commit_pending_transaction=False):
        wrapped_refcount = self._store_cache.check_in(self._store_cache_wrapped_key)
        if wrapped_refcount == 0:
            self.wrapped.close(commit_pending_transaction=commit_pending_transaction)
        else:
            L.debug("Wrapped store is still referenced %d times in the cache as %s. Cannot close it yet",
                    wrapped_refcount, self._store_cache_wrapped_key)

    def triples(self, pattern, context=None):
        ctxid = getattr(context, 'identifier', context)
        if ctxid in self.excludes:
            return
        for triple, contexts in self.wrapped.triples(pattern, context=context):
            filtered_contexts = self._contexts_filter(contexts)
            has_valid_contexts = next(filtered_contexts)
            if has_valid_contexts:
                yield triple, filtered_contexts

    def triples_choices(self, pattern, context=None):
        ctxid = getattr(context, 'identifier', context)
        if ctxid in self.excludes:
            return
        for triple, contexts in self.wrapped.triples_choices(pattern, context=context):
            filtered_contexts = self._contexts_filter(contexts)
            has_valid_contexts = next(filtered_contexts)
            if has_valid_contexts:
                yield triple, filtered_contexts

    def __len__(self, context=None):
        ctxid = getattr(context, 'identifier', context)
        if ctxid in self.excludes:
            return 0
        return sum(1 for _ in self.triples((None, None, None), context=context))

    def contexts(self, triple=None):
        cgen = self._contexts_filter(self.wrapped.contexts(triple))
        next(cgen)
        for c in cgen:
            yield c

    def _contexts_filter(self, contexts):
        contexts_iter = iter(contexts)
        excludes = self.excludes
        for c in contexts_iter:
            ctxid = getattr(c, 'identifier', c)
            if ctxid not in excludes:
                yield True
                yield c
                break
        else:  # no break
            yield False
            return
        for c in contexts_iter:
            ctxid = getattr(c, 'identifier', c)
            if ctxid not in excludes:
                yield c

    def prefix(self, namespace):
        return self.wrapped.prefix(namespace)

    def namespace(self, prefix):
        return self.wrapped.namespace(prefix)

    def gc(self):
        self.wrapped.gc()

    def add(self, *args, **kwargs):
        raise NotImplementedError()

    def addN(self, *args, **kwargs):
        raise NotImplementedError()

    def remove(self, *args, **kwargs):
        raise NotImplementedError()

    def add_graph(self, *args, **kwargs):
        raise NotImplementedError()

    def remove_graph(self, *args, **kwargs):
        raise NotImplementedError()

    def create(self, *args, **kwargs):
        raise NotImplementedError()

    def destroy(self, *args, **kwargs):
        raise NotImplementedError()

    def commit(self, *args, **kwargs):
        raise NotImplementedError()

    def rollback(self, *args, **kwargs):
        raise NotImplementedError()

    def __repr__(self):
        return '%s(%s)' % (FCN(type(self)), repr(self.wrapped))


def _is_cacheable(store_key, store_conf):
    '''
    Determine whether the store store's key and configuration indicate that the
    corresponding store can be reused by another `BundleDependencyStore` rather than being
    created anew.

    Parameters
    ----------
    store_key : str
        The key by which the store type can be looked up as an RDFLib store plugin
    store_conf : object
        Configuration parameters, such as would be passed to `Store.open`, which may
        indicate whether the store can be cached

    Returns
    -------
    bool
        True If the given store configuration is cacheable
    '''
    if store_key == 'agg':
        return all(_is_cacheable(k, c) for k, c in store_conf)
    if store_key == 'FileStorageZODB':
        if isinstance(store_conf, dict) and store_conf.get('read_only', False):
            return True
    if store_key == RDFLIB_PLUGIN_KEY:
        if isinstance(store_conf, (list, tuple)):
            try:
                kind, conf = store_conf
            except ValueError:
                L.warning('Inappropriate configuration for ' + RDFLIB_PLUGIN_KEY)
                return False
        elif isinstance(store_conf, dict):
            try:
                kind = store_conf['type']
                conf = store_conf['conf']
            except KeyError:
                L.warning('Inappropriate configuration for ' + RDFLIB_PLUGIN_KEY)
                return False
        else:
            L.warning('Unknown type of configuration for ' + RDFLIB_PLUGIN_KEY)
            return False
        return _is_cacheable(kind, conf)
    return False


def _cache_key(store_key, store_conf):
    '''
    Produces a key for use in the store cache.

    Parameters
    ----------
    store_key : str
        The key for the type of store which would be cached
    store_conf : object
        The configuration parameters for the store which would be cached
    '''
    if store_key == RDFLIB_PLUGIN_KEY and isinstance(store_conf, dict):
        store_conf = dict(**store_conf)
        del store_conf['cache']

    return dumps([store_key, store_conf],
            separators=(',', ':'),
            sort_keys=True)


class StoreCache(object):
    '''
    Cache of stores previously cached by a `BDS <BundleDependencyStore>`.

    We don't want to keep hold of a store if there's no BDS using it, so we only reference
    the stores weakly.
    '''
    def __init__(self):
        self._cache = WeakValueDictionary()

        self._refcounts = dict()
        '''
        Counts for references to cached BDS stores. Needed so we know when we can do clean-up.
        '''

    def check_out(self, key):
        self._refcounts[key] = self._refcounts.get(key, 0) + 1

    def check_in(self, key):
        rc = self._refcounts[key]
        if rc == 1:
            del self._refcounts[key]
        else:
            self._refcounts[key] = rc - 1
        return rc - 1

    def refcount(self, key):
        return self._refcounts.get(key, 0)

    def get(self, key, default=None):
        return self._cache.get(key, default)

    def __getitem__(self, key):
        return self._cache[key]

    def __setitem__(self, key, val):
        if key in self._cache:
            raise Exception(f"There's already an entry for {key}")
        self._cache[key] = val



from . import BASE_CONTEXT
from .dataobject import DataObject, ObjectProperty, This
from .context_common import CONTEXT_IMPORTS


# ContextDataObject was moved to a separate module from Context to avoid a dependency
# cycle with dataobject

class ContextDataObject(DataObject):
    """ Represents a context """
    class_context = BASE_CONTEXT
    rdf_type = BASE_CONTEXT.base_namespace['Context']
    imports = ObjectProperty(value_type=This,
                             multiple=True,
                             link=CONTEXT_IMPORTS)



from __future__ import print_function
import hashlib

from rdflib.term import URIRef
from urllib.parse import quote
from six import string_types

from .graph_object import IdentifierMissingException


__all__ = ['IdMixin']


# This was pulled out of DataObject when it was used by multiple classes. It may prove
# useful for making custom GraphObjects, so it's retained here as a mixin


class IdMixin:
    '''
    Mixin that provides common identifier logic

    Attributes
    ----------
    hashfun : function
        The function to use for encoding data provided to make_identifier.
        Should return an object can ``.encode()`` to a :py:class:`bytes` (a.k.a.
        :py:class:`str` in Python 2).  Defaults to :py:func:`hashlib.sha224`
    rdf_namespace : rdflib.namespace.Namespace
        The namespace for identifiers created
    direct_key : bool
        Whether to make a key directly, just adding the string onto the namespace or
        indirectly by hashing the key before joining with the namespace.
    '''
    hashfun = hashlib.sha224

    direct_key = True

    def __init__(self, ident=None, key=None, *args, **kwargs):
        super(IdMixin, self).__init__(*args, **kwargs)
        if key is not None and ident is not None:
            raise Exception("Only one of 'key' or 'ident' can be given to Context")

        if ident is not None:
            self._id = URIRef(ident)
        else:
            self._id = None

        self.key = key
        self._id_key = None
        self._generated_id = None

    @classmethod
    def make_identifier(cls, data):
        '''
        Makes an identifier based on this class' `rdf_namespace` by calling
        `__str__` on the data and passing to the class' `hashfun`.

        If the `__str__` for data's type doesn't function as an identifier, you
        should use either :meth:`make_identifier_direct` or override
        :meth:`identifier_augment` and :meth:`defined_augment`
        '''
        strdata = str(data)
        if strdata:
            hsh = "a" + cls.hashfun(strdata.encode()).hexdigest()
            return URIRef(cls.rdf_namespace[hsh])
        else:
            raise ValueError('Cannot use falsy value'
                             ' {} to make an identifier'.format(strdata))

    @classmethod
    def make_identifier_direct(cls, string):
        '''
        Make identifier by using the `~urllib.parse.quote`'d value of `key` appended to
        the `rdf_namespace` value
        '''
        if not isinstance(string, string_types):
            raise ValueError('make_identifier_direct only accepts strings')
        return URIRef(cls.rdf_namespace[quote(string)])

    def _gen_identifier(self):
        key = self.key
        if key is None:
            return None

        if key == self._id_key:
            return self._generated_id

        if self.direct_key:
            self._generated_id = self.make_identifier_direct(key)
        else:
            self._generated_id = self.make_identifier(key)
        self._id_key = key

        return self._generated_id

    @property
    def identifier(self):
        '''
        The identifier
        '''
        if self._id is not None:
            return self._id
        else:
            ident = self._gen_identifier()
            if ident is not None:
                return ident
            elif self.defined_augment():
                return self.identifier_augment()
            else:
                raise IdentifierMissingException(self)

    @identifier.setter
    def identifier(self, value):
        self._id = value

    def identifier_augment(self):
        """
        Override this method to define an identifier in lieu of one explicity set.

        One must also override :meth:`defined_augment` to return True whenever
        this method could return a valid identifier.
        :exc:`~.graph_object.IdentifierMissingException` should be
        raised if an identifier cannot be generated by this method.

        Raises
        ------
        `~.graph_object.IdentifierMissingException`

        """
        raise IdentifierMissingException(self)

    @property
    def defined(self):
        if self._id is not None or self.key is not None:
            return True
        else:
            return self.defined_augment()

    def defined_augment(self):
        """ This fuction must return False if :meth:`identifier_augment` would
        raise an :exc:`~.graph_object.IdentifierMissingException`. Override
        it when defining a non-standard identifier for subclasses of DataObjects.
        """
        return False



import http.client
import io
import logging
import os
from os.path import join as p, expanduser
import ssl
from urllib.parse import quote as urlquote, urlparse
import hashlib
import json
import pickle

from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache
import requests

from ...command_util import GenericUserError
from ...utils import FCN, retrieve_provider

from .. import URLConfig
from ..archive import ensure_archive, Unarchiver
from ..common import BUNDLE_ARCHIVE_MIME_TYPE

from . import LoadFailed, Loader, Uploader


L = logging.getLogger(__name__)


class HTTPURLConfig(URLConfig):
    '''
    HTTP URL configuration
    '''

    def __init__(self, *args,
            session_file_name=None,
            session_provider=None,
            cache_dir=None,
            mem_cache=False,
            **kwargs):
        '''
        Parameters
        ----------
        *args
            Passed on to URLConfig
        session_file_name : str, optional
            Session file name
        session_provider : str, optional
            Provider path for a callable that returns a session
        cache_dir : str, optional
            HTTP cache directory. Supersedes `mem_cache`
        mem_cache : bool, optional
            Whether to use an in-memory cache. Superseded by `cache_dir`
        **kwargs
            Passed on to URLConfig
        '''
        super(HTTPURLConfig, self).__init__(*args, **kwargs)
        self.cache_dir = cache_dir
        self.session_file_name = session_file_name
        self.session_provider = session_provider
        self.mem_cache = bool(mem_cache)
        self._session = None

    @property
    def session(self):
        '''
        A `requests.Session`

        This will be loaded from `.session_file_name` if a value is set for that.
        Otherwise, the session will either be obtained from the `.session_provider` or a
        default session will be created; in either case, any response caching
        configuration will be applied.
        '''
        if self._session is None:
            if self.session_file_name:
                try:
                    with open(expanduser(self.session_file_name), 'rb') as session_file:
                        self._session = pickle.load(session_file)
                except FileNotFoundError:
                    pass

            if self._session is None:
                self.init_session()

        return self._session

    def init_session(self):
        '''
        Initialize the HTTP session. Typically you won't call this, but will just access
        `.session`
        '''
        self._session = self._make_new_session()

    def _make_new_session(self):
        if self.session_provider:
            session = self._provide_session()
        else:
            session = requests.Session()

        if self.cache_dir:
            http_cache = FileCache(self.cache_dir)
            return CacheControl(session, cache=http_cache)
        elif self.mem_cache:
            return CacheControl(session)
        else:
            return session

    def _provide_session(self):
        return retrieve_provider(self.session_provider)()

    def save_session(self):
        sfname = expanduser(self.session_file_name)
        with open(sfname + '.tmp', 'wb') as session_file:
            pickle.dump(self._session, session_file)
        os.rename(sfname + '.tmp', sfname)

    def __getstate__(self):
        state = self.__dict__.copy()
        # We're storing the session in a separate file (if at all), so obviously we don't
        # to persist it with the HTTPURLConfig
        del state['_session']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._session = None


class HTTPSURLConfig(HTTPURLConfig):
    '''
    HTTPS URL configuration
    '''

    def __init__(self, *args, ssl_context_provider=None,
            ssl_context=None, **kwargs):
        '''
        Parameters
        ----------
        *args
            Passed on to HTTPURLConfig
        ssl_context_provider : str
            Path to a callable that provides a `ssl.SSLContext`. See `https_remote`
        ssl_context : ssl.SSLContext
            The SSL/TLS context to use for uploading with this accessor
        **kwargs
            Passed on to HTTPURLConfig
        '''
        super(HTTPSURLConfig, self).__init__(*args, **kwargs)
        self.ssl_context_provider = ssl_context_provider
        self._ssl_context = ssl_context

    def init_ssl_context(self):
        if self._ssl_context is not None:
            return

        if self.ssl_context_provider:
            ssl_context_provider = self._lookup_ssl_context_provider()

            try:
                ssl_context = ssl_context_provider()
            except Exception as e:
                raise HTTPSURLError('Error from SSL context provider'
                        f' "{self.ssl_context_provider}": {e}')

            if not isinstance(ssl_context, ssl.SSLContext):
                raise HTTPSURLError('Provider returned something other than an'
                        f' ssl.SSLContext: {ssl_context}')

            self._ssl_context = ssl_context

    def _lookup_ssl_context_provider(self):
        try:
            return retrieve_provider(self.ssl_context_provider)
        except ValueError:
            raise HTTPSURLError('Format of the provider path is incorrect')
        except AttributeError:
            raise HTTPSURLError(f'"{self.ssl_context_provider}" does not point to an'
                    ' SSL context provider')

    @property
    def ssl_context(self):
        self.init_ssl_context()
        return self._ssl_context

    def __getstate__(self):
        state = super(HTTPSURLConfig, self).__getstate__()
        del state['_ssl_context']
        return state

    def __setstate__(self, state):
        super(HTTPSURLConfig, self).__setstate__(state)
        self._ssl_context = None

    def __str__(self):
        if self.ssl_context_provider:
            ssl_context_maybe = f'\n    SSL Context Provider: {self.ssl_context_provider}'
        else:
            ssl_context_maybe = ''
        return f'{self.url}{ssl_context_maybe}'


HTTPSURLConfig.register('https')
HTTPURLConfig.register('http')


class HTTPSURLError(Exception):
    pass


class HTTPBundleLoader(Loader):
    '''
    Loads bundles from HTTP(S) resources listed in an index file
    '''

    def __init__(self, index_url, cachedir=None, hash_preference=('sha224',), **kwargs):
        '''
        Parameters
        ----------
        index_url : str or owmeta_core.bundle.URLConfig
            URL for the index file pointing to the bundle archives
        cachedir : str, optional
            Directory where the index and any downloaded bundle archive should be cached.
            If provided, the index and bundle archive is cached in the given directory. If
            not provided, the index will be cached in memory and the bundle will not be
            cached.
        hash_preference : tuple of str
            Preference ordering of hashes to use for checking integrity of files. If none
            match in the preference ordering, then the first one
        **kwargs
            Passed on to `.Loader`
        '''
        super(HTTPBundleLoader, self).__init__(**kwargs)

        if isinstance(index_url, str):
            self.index_url = index_url
            self._url_config = None
        elif isinstance(index_url, HTTPURLConfig):
            self.index_url = index_url.url
            self._url_config = index_url
        elif isinstance(index_url, URLConfig):
            self.index_url = index_url.url
            self._url_config = None
        else:
            raise TypeError('Expecting a string or URLConfig. Received %s' %
                    type(index_url))

        if not hash_preference:
            hash_preference = tuple(hash_preference)

        for hash_name in hash_preference:
            if hash_name not in hashlib.algorithms_available:
                raise ValueError(f'"{hash_name}" is not available in hashlib on this system')

        self.hash_preference = hash_preference
        self.cachedir = cachedir
        self._session = getattr(index_url, 'session', None) or requests.Session()
        self._index = None

    def __repr__(self):
        return '{}({})'.format(FCN(type(self)), repr(self.index_url))

    def _setup_index(self):
        if self._index is None:
            response = self._session.get(self.index_url)
            if response.status_code != 200:
                raise IndexLoadFailed(response)
            try:
                self._index = response.json()
            except json.decoder.JSONDecodeError:
                raise IndexLoadFailed(response)

    @classmethod
    def can_load_from(cls, ac):
        '''
        Returns `True` for ``http://`` or ``https://`` `URLConfigs <URLConfig>`

        Parameters
        ----------
        ac : AccessorConfig
            The config which we may be able to load from
        '''
        return (isinstance(ac, URLConfig) and
                (ac.url.startswith('https://') or
                    ac.url.startswith('http://')))

    def _save_session(self):
        if not self._url_config:
            return

        try:
            self._url_config.save_session()
        except Exception:
            L.warning('Error while attempting to save session', exc_info=True)

    def can_load(self, bundle_id, bundle_version=None):
        '''
        Check the index for an entry for the bundle.

        - If a version is given and the index has an entry for the bundle at that version
          and that entry gives a URL for the bundle, then we return `True`.

        - If no version is given and the index has an entry for the bundle at any version
          and that entry gives a URL for the bundle, then we return `True`.

        - Otherwise, we return `False`

        Parameters
        ----------
        bundle_id : str
            ID of the bundle to look for
        bundle_version : int, optional
            Version number of the bundle to look for. If not provided, then any version is
            deemed acceptable

        Returns
        -------
        bool
            `True` if the bundle can be loaded; otherwise, `False`
        '''
        try:
            self._setup_index()
            self._save_session()
        except IndexLoadFailed:
            L.warn('Failed to set up the index for %s', self,
                exc_info=L.isEnabledFor(logging.DEBUG))
            return False
        binfo = self._index.get(bundle_id)
        if binfo:
            if bundle_version is None:
                for binfo_version, versioned_binfo in binfo.items():
                    try:
                        int(binfo_version)
                    except ValueError:
                        L.warning("Got unexpected non-version-number key '%s' in bundle index info", binfo_version)
                        continue
                    try:
                        binfo_url = versioned_binfo.get('url')
                    except AttributeError:
                        L.warning("Got unexpected bundle info for version '%s' in bundle index info", binfo_version)
                        continue

                    if self._bundle_url_is_ok(binfo_url):
                        return True
                return False
            if not isinstance(binfo, dict):
                return False

            versioned_binfo = binfo.get(str(bundle_version))
            try:
                binfo_url = versioned_binfo.get('url')
            except AttributeError:
                L.warning("Got unexpected bundle info for version '%s' in bundle index info", versioned_binfo)
                return False
            return self._bundle_url_is_ok(binfo_url)

    def _bundle_url_is_ok(self, bundle_url):
        try:
            parsed_url = urlparse(bundle_url)
        except Exception:
            L.warning("Failed while parsing bundle URL '%s'", bundle_url)
            return False
        if parsed_url.scheme in ('http', 'https') and parsed_url.netloc:
            return True
        return False

    def bundle_versions(self, bundle_id):
        self._setup_index()
        binfo = self._index.get(bundle_id)

        if not binfo:
            return []

        res = []
        for k in binfo.keys():
            try:
                val = int(k)
            except ValueError:
                L.warning("Got unexpected non-version-number key '%s' in bundle index info", k)
            else:
                res.append(val)
        return res

    def load(self, bundle_id, bundle_version=None):
        try:
            return self._load(bundle_id, bundle_version)
        finally:
            self._save_session()

    def _load(self, bundle_id, bundle_version=None):
        '''
        Loads a bundle by downloading an index file, looking up the bundle location, and
        then downloading the bundle
        '''
        self._setup_index()
        binfo = self._index.get(bundle_id)
        if not binfo:
            raise LoadFailed(bundle_id, self, 'Bundle is not in the index')
        if not isinstance(binfo, dict):
            raise LoadFailed(bundle_id, self, 'Unexpected type of bundle info in the index')

        if bundle_version is None:
            max_vn = 0
            for k in binfo.keys():
                try:
                    val = int(k)
                except ValueError:
                    L.warning("Got unexpected non-version-number key '%s' in bundle index info", k)
                else:
                    if max_vn < val:
                        max_vn = val
            if not max_vn:
                raise LoadFailed(bundle_id, self, 'No releases found')
            bundle_version = max_vn

        versioned_binfo = binfo.get(str(bundle_version))

        if not versioned_binfo or not isinstance(versioned_binfo, dict):
            raise LoadFailed(bundle_id, self, f'No bundle info for version {bundle_version}')

        bundle_url = versioned_binfo.get('url')

        if not self._bundle_url_is_ok(bundle_url):
            raise LoadFailed(bundle_id, self, 'Did not find a valid URL for "%s" at'
                    ' version %s' % (bundle_id, bundle_version))

        hashes = versioned_binfo.get('hashes')
        if not isinstance(hashes, dict) or not hashes:
            raise LoadFailed(bundle_id, self, f'No hash info for version {bundle_version}')

        for hash_name in self.hash_preference:
            bundle_hash = hashes.get(hash_name)
            if bundle_hash:
                break
        else: # no break
            for hash_name, bundle_hash in hashes.items():
                if hash_name in hashlib.algorithms_available:
                    break
            else: # no break
                raise LoadFailed(bundle_id, self, f'No supported hash for version {bundle_version}')

        try:
            hsh = hashlib.new(hash_name)
        except ValueError:
            L.warning('Hash in hashlib.algorithms_available unsupported in hashlib.new')
            raise LoadFailed(bundle_id, self, f'Unsupported hash {hash_name} for version {bundle_version}')

        response = self._session.get(bundle_url, stream=True)
        if self.cachedir is not None:
            bfn = urlquote(bundle_id)
            with open(p(self.cachedir, bfn), 'wb') as f:
                for chunk in response.iter_content(chunk_size=1024):
                    hsh.update(chunk)
                    f.write(chunk)
            if bundle_hash != hsh.hexdigest():
                raise LoadFailed(bundle_id, self,
                        f'Failed to verify {hash_name} hash for version {bundle_version}')
            with open(p(self.cachedir, bfn), 'rb') as f:
                Unarchiver().unpack(f, self.base_directory)
        else:
            bio = io.BytesIO()
            bundle_bytes = response.raw.read()
            hsh.update(bundle_bytes)
            if bundle_hash != hsh.hexdigest():
                raise LoadFailed(bundle_id, self,
                        f'Failed to verify {hash_name} hash for version {bundle_version}')
            bio.write(bundle_bytes)
            bio.seek(0)
            Unarchiver().unpack(bio, self.base_directory)


class HTTPBundleUploader(Uploader):
    '''
    Uploads bundles by sending bundle archives in HTTP POST requests
    '''

    def __init__(self, upload_url, ssl_context=None):
        '''
        Parameters
        ----------
        upload_url : str or URLConfig
            URL string or accessor config
        ssl_context : ssl.SSLContext
            SSL/TLS context to use for the connection. Overrides any context provided in
            `upload_url`
        '''
        super(HTTPBundleUploader, self).__init__()

        self.ssl_context = None

        if isinstance(upload_url, str):
            self.upload_url = upload_url
        elif isinstance(upload_url, HTTPSURLConfig):
            self.upload_url = upload_url.url
            self.ssl_context = upload_url.ssl_context
        elif isinstance(upload_url, URLConfig):
            self.upload_url = upload_url.url
        else:
            raise TypeError('Expecting a string or URLConfig. Received %s' %
                    type(upload_url))

        if ssl_context:
            self.ssl_context = ssl_context

    @classmethod
    def can_upload_to(self, accessor_config):
        return (isinstance(accessor_config, URLConfig) and
                (accessor_config.url.startswith('https://') or
                    accessor_config.url.startswith('http://')))

    def upload(self, bundle_path):
        with ensure_archive(bundle_path) as archive_path:
            self._post(archive_path)

    def _post(self, archive):
        parsed_url = urlparse(self.upload_url)
        if parsed_url.scheme == 'http':
            connection_ctor = http.client.HTTPConnection
        else:
            def connection_ctor(*args, **kwargs):
                return http.client.HTTPSConnection(*args,
                        context=self.ssl_context, **kwargs)
        conn = connection_ctor(parsed_url.netloc)
        with open(archive, 'rb') as f:
            conn.request("POST", "", body=f, headers={'Content-Type':
                BUNDLE_ARCHIVE_MIME_TYPE})
        # XXX: Do something with this response
        # conn.getresponse()


def http_remote(self, *, cache=None, session_provider=None, session_file_name=None):
    '''
    Provide additional parameters for HTTP remote accessors

    Parameters
    ----------
    cache : str
        Either the string "mem" or a file path to a cache directory
    session_provider : str
        Path to a callable that provides a `requests.Session`. The format is similar to
        that for setuptools entry points: ``path.to.module:path.to.provider.callable``.
        Notably, there's no name and "extras" are not supported. optional.
    session_file_name : str
        Path to a file where the HTTP session can be stored
    '''

    if self._url_config is None:
        raise GenericUserError('An HTTP URL must be specified for HTTP accessors')

    if not isinstance(self._url_config, HTTPURLConfig):
        raise GenericUserError(f'The specified URL, {self._url_config} is not an HTTP URL')

    _http_urlconfig_command_helper(self, cache, session_provider, session_file_name)

    return self._write_remote()


def https_remote(self, *, ssl_context_provider=None, cache=None, session_provider=None,
        session_file_name=None):
    '''
    Provide additional parameters for HTTPS remote accessors

    Parameters
    ----------
    ssl_context_provider : str
        Path to a callable that provides a `ssl.SSLContext` used for bundle uploads. The
        format is similar to that for setuptools entry points:
        ``path.to.module:path.to.provider.callable``.  Notably, there's no name and
        "extras" are not supported. optional.
    cache : str
        Either the string "mem" or a file path to a cache directory
    session_provider : str
        Path to a callable that provides a `requests.Session`. The format is similar to
        that for setuptools entry points: ``path.to.module:path.to.provider.callable``.
        Notably, there's no name and "extras" are not supported. optional.
    session_file_name : str
        Path to a file where the HTTP session can be stored
    '''

    if self._url_config is None:
        raise GenericUserError('An HTTPS URL must be specified for HTTPS accessors')

    if not isinstance(self._url_config, HTTPSURLConfig):
        raise GenericUserError(f'The specified URL, {self._url_config} is not an HTTPS URL')

    try:
        if ssl_context_provider:
            self._url_config.ssl_context_provider = ssl_context_provider
            self._url_config.init_ssl_context()
    except HTTPSURLError as e:
        raise GenericUserError(str(e))

    _http_urlconfig_command_helper(self, cache, session_provider, session_file_name)

    return self._write_remote()


def _http_urlconfig_command_helper(self, cache, session_provider, session_file_name):
    if cache == 'mem':
        self._url_config.mem_cache = True
    else:
        self._url_config.cache_dir = cache

    if session_file_name:
        self._url_config.session_file_name = session_file_name

    if session_provider:
        self._url_config.session_provider = session_provider

    # Initialize a session to make sure the configs work.
    self._url_config._make_new_session()


class IndexLoadFailed(Exception):
    '''
    Thrown when the HTTP bundle loader cannot get its index
    '''
    def __init__(self, response):
        self.response = response



def hash_file(hsh, fname, blocksize=None):
    '''
    Updates the given hash object with the contents of a file.

    The file is read in `blocksize` chunks to avoid eating up too much memory at a time.

    Parameters
    ----------
    hsh : `hashlib.hash <hashlib>`
        The hash object to update
    fname : str
        The filename for the file to hash
    blocksize : int, optional
        The number of bytes to read at a time. If not provided, will use
        `hsh.block_size <hashlib.hash.block_size>` instead.
    '''
    if not blocksize:
        blocksize = hsh.block_size

    with open(fname, 'rb') as fh:
        while True:
            block = fh.read(blocksize)
            if not block:
                break
            hsh.update(block)



from __future__ import print_function
from __future__ import absolute_import

from collections import OrderedDict, defaultdict
import logging

from rdflib.term import URIRef
import six

from . import BASE_CONTEXT
from .utils import FCN
from .context import Context
from .dataobject import (DataObject, ObjectProperty, DatatypeProperty, UnionProperty, This,
                         CPThunk)
from .data_trans.common_data import DS_NS, DS_DATA_NS
from .graph_object import IdentifierMissingException

L = logging.getLogger(__name__)

INFO_PROP_PREFIX = '_info_prop_'


class FormatUtil(object):
    @staticmethod
    def collect_values(attr, stored):
        if stored:
            attr_vals = list()
            for x in attr.get():
                if x not in attr_vals:
                    attr_vals.append(x)
        else:
            attr_vals = attr.defined_values
        return attr_vals


class Informational(object):
    '''
    Defines a property on a `.DataSource`

    Attributes
    ----------
    name : str
        The name for the property
    description : str
        A description of the property
    also : tuple of Informational or list of Informational
        Other properties which, if set, set the value for this property. If multiple such
        "also" properties are set when the owning `DataSource` instance is defined, then
        a `DuplicateAlsoException` will be raised.
    default_override : object
        An override for the default value, typically set by setting the value in a
        `.DataSource` class dictionary. Importantly, this overrides an "also" value which
        would normally take precedence.
    default_value : object
        Default value if no other value is set
    multiple : boolean
        If `True`, then the property can take on multiple values for the same subject
    cls : type
        The `~owmeta_core.dataobject_property.Property` corresponding to this property
    '''

    def __init__(self, name=None, display_name=None, description=None,
                 default_value=None, property_type='DatatypeProperty',
                 multiple=True, property_name=None, also=(), **property_args):
        '''
        Parameters
        ----------
        name : str, optional
            Name for the property. If not provided here, then the name generally gets set
            to the name to which this object is assigned
        display_name : str, optional
            Display name for the property. If not provided here, then the `name` will be
            used for the display name
        description : str, optional
            A description of the property
        default_value : object, optional
            Value to use
        property_type : 'DatatypeProperty', 'ObjectProperty', or 'UnionProperty'
            The type of `~owmeta_core.dataobject_property.Property` to create from this object.
            Default is 'DatatypeProperty'
        multiple : boolean, optional
            Whether this property can have multiple values for the same object. Default is
            `True`
        property_name : str, optional
            The name of the property to use for attributes. `name` will be used if a value
            is not provided here
        also : Informational, tuple of Informational, or list of Informational; optional
            Other properties which, if set, will give their value to this property as well
        **property_args
            Additional arguments which will be passed into the class dictionary when the
            `~owmeta_core.dataobject_property.Property` corresponding to this object is created.
        '''
        self.name = name
        self._property_name = property_name
        self._display_name = display_name
        self.default_value = default_value
        self.description = description
        self.property_type = property_type
        self.multiple = multiple
        if also and not isinstance(also, (list, tuple)):
            also = (also,)
        self.also = also
        self.property_args = property_args

        self.default_override = None

        self.cls = None

    def __get__(self, obj, owner):
        if obj is None:
            return self
        else:
            return getattr(obj, INFO_PROP_PREFIX + self.name)

    @property
    def display_name(self):
        '''
        The display name for the property.
        '''
        return self._display_name if self._display_name is not None else self.name

    @display_name.setter
    def display_name(self, val):
        self._display_name = val

    @property
    def property_name(self):
        '''
        The name of the property to use for attributes
        '''
        return self._property_name if self._property_name is not None else self.name

    @property_name.setter
    def property_name(self, v):
        self._property_name = v

    def copy(self):
        '''
        Copy to a new `Informational`
        '''
        res = type(self)()
        for x in vars(self):
            setattr(res, x, getattr(self, x))
        return res

    def __repr__(self):
        return ("Informational(name='{}',"
                " display_name={},"
                " default_value={},"
                " description={})").format(self.name,
                                          repr(self.display_name),
                                          repr(self.default_value),
                                          repr(self.description))

    # NOTE: This guy has to come last to avoid conflict with the decorator
    @property
    def property(self):
        return getattr(self.cls, INFO_PROP_PREFIX + self.name).property


class DuplicateAlsoException(Exception):
    pass


class DataSourceType(type(DataObject)):
    """A type for DataSources

    Sets up the graph with things needed for MappedClasses
    """

    def __init__(self, name, bases, dct):
        self.__info_fields = []
        others = []
        newdct = dict()
        for z in dct:
            meta = dct[z]
            if isinstance(meta, Informational):
                if meta.cls is not None:
                    L.debug("Already created a Property from %s for %s. Not creating another for %s",
                            meta, meta.cls, self)
                    meta_owner_property_property = getattr(meta.cls, INFO_PROP_PREFIX + meta.name)
                    newdct[INFO_PROP_PREFIX + z] = CPThunk(meta_owner_property_property.property)

                    meta_copy = meta.copy()
                    meta_copy.cls = self
                    meta_copy.name = z
                    self.__info_fields.append(meta_copy)
                    setattr(self, z, meta_copy)
                else:
                    meta.cls = self
                    meta.name = z
                    self.__info_fields.append(meta)

                    # Make the owmeta_core property
                    #
                    # We set the name for the property to the inf.name since that's how we
                    # access the info on this object, but the inf.property_name is used for
                    # the linkName so that the property's URI is generated based on that name.
                    # This allows to set an attribute named inf.property_name on self while
                    # still having access to the property through inf.name.
                    ptype = None
                    if meta.property_type == 'DatatypeProperty':
                        ptype = DatatypeProperty
                    elif meta.property_type == 'ObjectProperty':
                        ptype = ObjectProperty
                    elif meta.property_type == 'UnionProperty':
                        ptype = UnionProperty
                    else:
                        raise ValueError(f'Unrecognized property type {meta.property_type}')

                    newdct[INFO_PROP_PREFIX + meta.name] = ptype(
                            linkName=meta.property_name,
                            multiple=meta.multiple,
                            **meta.property_args)
            else:
                others.append((z, dct[z]))

        for x in bases:
            if isinstance(x, DataSourceType):
                self.__info_fields += [inf.copy() for inf in x.__info_fields]

        for k, v in others:
            for i in range(len(self.__info_fields)):
                if self.__info_fields[i].name == k:
                    # This is for setting default values from a super-class. We copy the
                    # Informational because the default is baked-in to the Informational
                    # instance, and we want it to apply only to the sub-class
                    self.__info_fields[i].default_override = v
                    setattr(self, k, self.__info_fields[i])
                    break
            else: # no 'break'
                newdct[k] = v
        if not getattr(self, '__doc__', None):
            self.__doc__ = self._docstr()
        super(DataSourceType, self).__init__(name, bases, newdct)

    def _docstr(self):
        s = ''
        for inf in self.__info_fields:
            s += '{} : :class:`~owmeta_core.dataobject.{}`'.format(inf.display_name, inf.property_type) + \
                    ('\n    Attribute: `{}`'.format(inf.name if inf.property_name is None else inf.property_name)) + \
                    (('\n\n    ' + inf.description) if inf.description else '') + \
                    ('\n\n    Default value: {}'.format(inf.default_value) if inf.default_value is not None else '') + \
                    '\n\n'
        return s

    @property
    def info_fields(self):
        return self.__info_fields


class DataSource(six.with_metaclass(DataSourceType, DataObject)):
    '''
    A source for data that can get translated into owmeta_core objects.

    The value for any field can be passed to __init__ by name. Additionally, if
    the sub-class definition of a DataSource assigns a value for that field like::

        class A(DataSource):
            some_field = 3

    that value will be used over the default value for the field, but not over
    any value provided to __init__.
    '''

    class_context = BASE_CONTEXT

    source = Informational(display_name='Input source',
                           description='The data source that was translated into this one',
                           property_type='ObjectProperty',
                           value_type=This)

    translation = Informational(display_name='Translation',
                                description='Information about the translation process that created this object',
                                property_type='ObjectProperty',
                                cascade_retract=True)

    description = Informational(display_name='Description',
                                description='Free-text describing the data source')

    base_namespace = DS_NS
    base_data_namespace = DS_DATA_NS

    def __init__(self, **kwargs):
        # There's a similar behavior in vanilla DataObject, but that doesn't have default
        # defaults and default-overrides. We don't pass the arguments up to DataObject so
        # the `properties_are_init_args` handling isn't used (whether
        # `properties_are_init_args` is True or False we get bad or incomplete behavior
        # when the property arguments are passed up)
        self.info_fields = OrderedDict((i.name, i) for i in self.__class__.info_fields)
        parent_kwargs = dict()
        new_kwargs = dict()
        for k, v in kwargs.items():
            if k not in self.info_fields:
                parent_kwargs[k] = v
            else:
                new_kwargs[k] = v
        super(DataSource, self).__init__(**parent_kwargs)
        vals = defaultdict(dict)
        for n, inf in self.info_fields.items():
            v = new_kwargs.get(n, None)
            if v is not None:
                vals[n]['i'] = v
            else:
                v = inf.default_value

            if inf.default_override is not None:
                vals[n]['e'] = inf.default_override

            vals[n]['d'] = inf.default_value

            for also in inf.also:
                if v is not None and vals[also.name].setdefault('a', v) != v:
                    raise DuplicateAlsoException('Only one also is allowed')

        for n, vl in vals.items():
            inf = self.info_fields[n]
            v = vl.get('i', vl.get('e', vl.get('a', vl['d'])))
            ctxd_prop = getattr(self, INFO_PROP_PREFIX + inf.name)
            if v is not None:
                ctxd_prop(v)

    def commit(self):
        '''
        Commit the data source *locally*

        This includes staging files such as they would be available for a translation. In general, a sub-class should
        implement :meth:`commit_augment` rather than this method, or at least call this method via super
        '''
        self.commit_augment()

    def commit_augment(self):
        pass

    def defined_augment(self):
        return self.translation.has_defined_value()

    def identifier_augment(self):
        return self.make_identifier(self.translation.defined_values[0].identifier.n3())

    def __str__(self):
        return self.format_str(False)

    def format_str(self, stored):
        try:
            sio = six.StringIO()
            print(self.__class__.__name__, end='', file=sio)
            if self.defined:
                ident = self.identifier
                if self.namespace_manager:
                    ident = self.namespace_manager.normalizeUri(ident)
                print(f'({ident})', file=sio)
            else:
                print(file=sio)
            for info in self.info_fields.values():
                attr = getattr(self, info.name)
                attr_vals = FormatUtil.collect_values(attr, stored)
                if attr_vals:
                    print('    ' + info.display_name, end=': ', file=sio)
                    for val in sorted(attr_vals):
                        val_line_sep = '\n      ' + ' ' * len(info.display_name)
                        if isinstance(val, DataSource):
                            valstr = val.format_str(stored)
                        elif isinstance(val, GenericTranslation):
                            valstr = val.format_str(stored)
                        elif isinstance(val, URIRef):
                            valstr = val.n3()
                        elif isinstance(val, six.string_types):
                            valstr = repr(val)
                        else:
                            valstr = str(val)
                        print(val_line_sep.join(valstr.split('\n')), end=' ', file=sio)
                    print(file=sio)
            return sio.getvalue()
        except AttributeError:
            res = super(DataSource, self).__str__()
            L.error('Failed while creating formatting string representation for %s', res)
            return res


class Translation(DataObject):
    """
    Representation of the method by which a DataSource was translated and
    the sources of that translation.  Unlike the 'source' field attached to
    DataSources, the Translation may distinguish different kinds of input
    source to a translation.
    """

    class_context = BASE_CONTEXT

    translator = ObjectProperty()

    def defined_augment(self):
        return self.translator.has_defined_value() and self.translator.onedef().defined

    def identifier_augment(self):
        return self.make_identifier(self.translator.onedef().identifier.n3())


class GenericTranslation(Translation):
    """
    A generic translation that just has sources in any order
    """

    class_context = BASE_CONTEXT

    source = ObjectProperty(multiple=True, value_rdf_type=DataSource.rdf_type)

    def defined_augment(self):
        return super(GenericTranslation, self).defined_augment() and \
                self.source.has_defined_value()

    def identifier_augment(self):
        data = super(GenericTranslation, self).identifier_augment().n3() + \
                "".join(sorted(x.identifier.n3() for x in self.source.defined_values))
        return self.make_identifier(data)

    def __str__(self):
        return self.format_str(False)

    def format_str(self, stored):
        sio = six.StringIO()
        print('{}({})'.format(self.__class__.__name__, self.idl), file=sio)
        sources_field_name = 'Sources: '
        print(sources_field_name, end='', file=sio)

        attr = self.source
        attr_vals = FormatUtil.collect_values(attr, stored)

        if attr_vals:
            val_line_sep = '\n' + len(sources_field_name) * ' '
            print(val_line_sep.join(val_line_sep.join(val.format_str(stored).split('\n'))
                                    for val in sorted(attr_vals)), file=sio)

        if stored:
            translator = self.translator.one()
        else:
            translator = self.translator.onedef()
        if translator is not None:
            field = "Translator: "
            s = ('\n' + len(field) * ' ').join(str(translator).split('\n'))
            print(field + s, file=sio)
        return sio.getvalue()


class DataObjectContextDataSource(DataSource):

    class_context = BASE_CONTEXT

    def __init__(self, context, **kwargs):
        super(DataObjectContextDataSource, self).__init__(**kwargs)
        if context is not None:
            self.context = context
        else:
            self.context = Context()


def format_types(typ):
    if isinstance(typ, OneOrMore):
        return ':class:`{}` (:class:`~{}`)'.format(FCN(OneOrMore), FCN(typ.source_type))
    elif isinstance(typ, type):
        return ':class:`~{}`'.format(FCN(typ))
    else:
        return ', '.join(':class:`~{}`'.format(FCN(x)) for x in typ)


class DataTransatorType(type(DataObject)):
    def __init__(self, name, bases, dct):
        super(DataTransatorType, self).__init__(name, bases, dct)

        if not getattr(self, '__doc__', None):
            self.__doc__ = '''Input type(s): {}\n
                              Output type(s): {}\n'''.format(format_types(self.input_type),
                                                             format_types(self.output_type))


class BaseDataTranslator(six.with_metaclass(DataTransatorType, DataObject)):
    '''
    Translates from a data source to owmeta_core objects

    Attributes
    ----------
    input_type : type or tuple of type
        Types of input to this `DataTranslator`. Should be sub-classes of `DataSource`
    output_type : type or tuple of type
        Types of output from this `DataTranslator`. Should be sub-classes of `DataSource`
    translation_type : type
        Type of the translation record produced as a side-effect of translating with this
        translator
    output_key : str
        The "key" for outputs from this translator. See `IdentifierMixin`
    '''

    class_context = BASE_CONTEXT

    input_type = DataSource
    output_type = DataSource
    translation_type = Translation

    def __call__(self, *args, **kwargs):
        self.output_key = kwargs.pop('output_key', None)
        self.output_identifier = kwargs.pop('output_identifier', None)
        try:
            return self.translate(*args, **kwargs)
        finally:
            self.output_key = None
            self.output_identifier = None

    def __str__(self):
        s = '''Input type(s): {}
               Output type(s): {}'''.format(self.input_type,
                                            self.output_type)
        return f'{FCN(type(self))}({self.idl})' + ': \n    ' + ('\n    '.join(x.strip() for x in s.split('\n')))

    def defined_augment(self):
        return True

    def identifier_augment(self):
        return self.make_identifier(type(self).rdf_type)

    def translate(self, *args, **kwargs):
        '''
        Notionally, this method takes a data source, which is translated into
        some other data source. There doesn't necessarily need to be an input
        data source.
        '''
        raise NotImplementedError

    def make_translation(self, sources=()):
        '''
        It's intended that implementations of DataTranslator will override this
        method to make custom Translations according with how different
        arguments to Translate are (or are not) distinguished.

        The actual properties of a Translation subclass must be defined within
        the 'translate' method
        '''
        return self.translation_type.contextualize(self.context)(translator=self)

    def make_new_output(self, sources, *args, **kwargs):
        trans = self.make_translation(sources)
        res = self.output_type.contextualize(self.context)(*args, translation=trans,
                                                           ident=self.output_identifier,
                                                           key=self.output_key, **kwargs)
        for s in sources:
            # XXX: Why are we calling contextualize here? Seems redundant?
            res.contextualize(self.context).source(s)

        return res


class OneOrMore(object):
    """
    Wrapper for :class:`DataTranslator` input :class:`DataSource` types indicating that
    one or more of the wrapped type must be provided to the translator
    """
    def __init__(self, source_type):
        self.source_type = source_type

    def __repr__(self):
        return FCN(type(self)) + '(' + repr(self.source_type) + ')'


class DataTranslator(BaseDataTranslator):
    """
    A specialization with the :class:`GenericTranslation` translation type that adds
    sources for the translation automatically when a new output is made
    """

    class_context = BASE_CONTEXT

    translation_type = GenericTranslation

    def make_translation(self, sources=()):
        res = super(DataTranslator, self).make_translation(sources)
        for s in sources:
            res.source(s)
        return res


class PersonDataTranslator(BaseDataTranslator):
    """
    A person who was responsible for carrying out the translation of a data source
    manually
    """

    class_context = BASE_CONTEXT

    person = ObjectProperty(multiple=True,
            __doc__='A person responsible for carrying out the translation.')

    # No translate impl is provided here since this is intended purely as a descriptive object



from rdflib import plugin
from rdflib.store import Store, NO_STORE

from .utils import FCN


class AggregateStore(Store):
    '''
    A read-only aggregate of RDFLib `stores <rdflib.store.Store>`
    '''

    context_aware = True
    '''
    Specified by RDFLib. Required to be `True` for `~rdflib.graph.ConjunctiveGraph` stores.

    Aggregated stores MUST be context-aware. This is enforced by :meth:`open`.
    '''

    graph_aware = True
    '''
    Specified by RDFLib. Required to be `True` for `~rdflib.graph.Dataset` stores.

    The first store must be graph-aware. This is enforced by :meth:`open`.
    '''

    # Unlike "awareness" attributes, checking for support of range queries is handled
    # after the store is open, so we don't care if we need to change it to `True` later.
    supports_range_queries = False

    def __init__(self, configuration=None, identifier=None, graph_aware=None):
        super(AggregateStore, self).__init__(configuration, identifier)
        self.__stores = []
        self.__bound_ns = dict()
        if graph_aware is not None:
            self.graph_aware = graph_aware

    @property
    def stores(self):
        return list(self.__stores)

    # -- Store methods -- #

    def open(self, configuration, create=True):
        '''
        Creates and opens all of the stores specified in the configuration

        Also checks for all aggregated stores to be `context_aware`
        '''
        if not isinstance(configuration, (tuple, list)):
            return NO_STORE
        self.__stores = []
        for store_key, store_conf in configuration:
            store = plugin.get(store_key, Store)()
            store.open(store_conf)
            self.__stores.append(store)
        assert self.__stores, 'At least one store configuration must be provided'
        assert all(x.context_aware for x in self.__stores), ('All aggregated stores must be'
                                                             ' context_aware')
        assert self.__stores[0].graph_aware, 'The first store must be graph_aware'
        self.supports_range_queries = all(getattr(x, 'supports_range_queries', False) for x in self.__stores)

    def triples(self, triple, context=None):
        for store in self.__stores:
            for trip in store.triples(triple, context=context):
                yield trip

    def triples_choices(self, triple, context=None):
        for store in self.__stores:
            for trip in store.triples_choices(triple, context=context):
                yield trip

    def __len__(self, context=None):
        # rdflib specifies a context argument for __len__, but how do you even pass that
        # argument to len?
        return sum(len(store) for store in self.__stores)

    def contexts(self, triple=None):
        for store in self.__stores:
            for ctx in store.contexts(triple):
                yield ctx

    def prefix(self, namespace):
        prefix = None
        for store in self.__stores:
            aprefix = store.prefix(namespace)
            if aprefix and prefix and aprefix != prefix:
                msg = 'multiple prefixes ({},{}) for namespace {}'.format(prefix, aprefix, namespace)
                raise AggregatedStoresConflict(msg)
            prefix = aprefix
        return prefix

    def namespace(self, prefix):
        namespace = None
        for store in self.__stores:
            anamespace = store.namespace(prefix)
            if anamespace and namespace and anamespace != namespace:
                msg = 'multiple namespaces ({},{}) for prefix {}'.format(namespace, anamespace, prefix)
                raise AggregatedStoresConflict(msg)
            namespace = anamespace
        if namespace is None:
            namespace = self.__bound_ns.get(prefix)
        return namespace

    def namespaces(self):
        for store in self.__stores:
            for ns in store.namespaces():
                yield ns
        for ns in self.__bound_ns.items():
            yield ns

    def bind(self, prefix, namespace):
        self.__bound_ns[prefix] = namespace

    def close(self, *args, **kwargs):
        for store in self.__stores:
            store.close(*args, **kwargs)

    def gc(self):
        for store in self.__stores:
            store.gc()

    def add(self, *args, **kwargs):
        return self.__stores[0].add(*args, **kwargs)

    def addN(self, *args, **kwargs):
        return self.__stores[0].addN(*args, **kwargs)

    def remove(self, *args, **kwargs):
        self.__stores[0].remove(*args, **kwargs)

    def add_graph(self, *args, **kwargs):
        if not self.graph_aware:
            super().add_graph(*args, **kwargs)
        else:
            self.__stores[0].add_graph(*args, **kwargs)

    def remove_graph(self, *args, **kwargs):
        if not self.graph_aware:
            super().remove_graph(*args, **kwargs)
        else:
            self.__stores[0].remove_graph(*args, **kwargs)

    def create(self, *args, **kwargs): raise UnsupportedAggregateOperation
    def destroy(self, *args, **kwargs): raise UnsupportedAggregateOperation

    def rollback(self, *args, **kwargs):
        return self.__stores[0].rollback(*args, **kwargs)

    def commit(self, *args, **kwargs):
        return self.__stores[0].commit(*args, **kwargs)

    def __repr__(self):
        return '%s(%s)' % (FCN(type(self)), ', '.join(repr(s) for s in self.__stores))


class UnsupportedAggregateOperation(Exception):
    '''
    Thrown for operations which modify a graph and hence are inappropriate for
    `AggregateStore`
    '''


class AggregatedStoresConflict(Exception):
    pass



'''
Package for uploaders and downloaders of bundles
'''
import logging
from pkg_resources import iter_entry_points, DistributionNotFound


L = logging.getLogger(__name__)

LOADERS_GROUP = 'owmeta_core.loaders'


class LoadFailed(Exception):
    '''
    Thrown when a bundle could not be downloaded
    '''
    def __init__(self, bundle_id, loader, *args):
        '''
        Parameters
        ----------
        bundle_id : str
            ID of the bundle on which a download was attempted
        loader : Loader
            The loader that attempted to download the bundle
        args[0] : str
            Explanation of why the download failed
        *args[1:]
            Passed on to `Exception`
        '''
        msg = args[0]
        mmsg = 'Failed to load {} bundle with loader {}{}'.format(
                bundle_id, loader, ': ' + msg if msg else '')
        super(LoadFailed, self).__init__(mmsg, *args[1:])


class Loader(object):
    '''
    Downloads bundles into the local index and caches them

    Note that a `Loader` is transient: it will be created when needed to download *one*
    bundle and then discarded. Any state that should be carried from request to request
    should be attached to an `AccessorConfig`

    Attributes
    ----------
    base_directory : str
        The path where the bundle archive should be unpacked
    '''

    def __init__(self):
        # The base directory
        self.base_directory = None

    @classmethod
    def can_load_from(cls, accessor_config):
        '''
        Returns `True` if the given `accessor_config` is a valid config for this loader

        Parameters
        ----------
        accessor_config : AccessorConfig
            The config which we may be able to load from
        '''
        return False

    def can_load(self, bundle_id, bundle_version=None):
        '''
        Returns True if the bundle named `bundle_id` is available.

        This method is for loaders to determine that they probably can or cannot load the
        bundle, such as by checking repository metadata. Other loaders that return `True`
        from `can_load` should be tried if a given loader fails, but a warning should be
        recorded for the loader that failed.
        '''
        return False

    def bundle_versions(self, bundle_id):
        '''
        List the versions available for the bundle.

        This is a required part of the `Loader` interface.

        Parameters
        ----------
        bundle_id : str
            ID of the bundle for which versions are requested

        Returns
        -------
        A list of int
            Each entry is a version of the bundle available via this loader
        '''
        raise NotImplementedError()

    def load(self, bundle_id, bundle_version=None):
        '''
        Load the bundle into the local index

        Parameters
        ----------
        bundle_id : str
            ID of the bundle to load
        bundle_version : int
            Version of the bundle to load. Defaults to the latest available. optional
        '''
        raise NotImplementedError()

    def __call__(self, bundle_id, bundle_version=None):
        '''
        Load the bundle into the local index. Short-hand for `load`
        '''
        return self.load(bundle_id, bundle_version)

    @classmethod
    def register(cls):
        LOADER_CLASSES.add(cls)


class Uploader(object):
    '''
    Uploads bundles to remotes
    '''

    @classmethod
    def can_upload_to(self, accessor_config):
        '''
        Returns True if this uploader can upload with the given accessor configuration

        Parameters
        ----------
        accessor_config : AccessorConfig
        '''
        return False

    def can_upload(self, bundle_path):
        '''
        Returns True if this uploader can upload this bundle

        Parameters
        ----------
        bundle_path : str
            The file path to the bundle to upload
        '''
        return False

    def upload(self, bundle_path):
        '''
        Upload a bundle

        Parameters
        ----------
        bundle_path : str
            The file path to the bundle to upload
        '''
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        return self.upload(*args, **kwargs)

    @classmethod
    def register(cls):
        UPLOADER_CLASSES.add(cls)


def load_entry_point_loaders():
    for entry_point in iter_entry_points(group=LOADERS_GROUP):
        try:
            entry_point.load().register()
        except DistributionNotFound:
            # This is expected...there's no pre-filtering of entry points for when extras
            # aren't installed...maybe to allow for optional depend on extras
            L.debug('Not adding loader %s due to failure in package resources resolution',
                    entry_point, exc_info=True)
        except Exception:
            L.warning('Unable to register loader %s', entry_point.name,
                    exc_info=L.isEnabledFor(logging.DEBUG))


LOADER_CLASSES = set()


UPLOADER_CLASSES = set()



from itertools import chain

from rdflib.store import Store, VALID_STORE, NO_STORE
from rdflib.plugins.memory import IOMemory
from rdflib.term import Variable

from .context_common import CONTEXT_IMPORTS
from .rdf_utils import transitive_lookup


class ContextStoreException(Exception):
    pass


class ContextStore(Store):
    '''
    A store specific to a `~owmeta_core.context.Context`


    A `ContextStore` may have triples
    '''

    context_aware = True

    def __init__(self, context=None, include_stored=False, imports_graph=None, **kwargs):
        """
        Parameters
        ----------
        context : ~owmeta_core.context.Context
            The context to which this store belongs
        include_stored : bool
            If `True`, the backing store will be queried as well as the staged triples in
            `context`
        imports_graph : ~rdflib.store.Store or ~rdflib.graph.Graph
            The graph to query for imports relationships between contexts
        **kwargs
            Passed on to `Store <rdflib.store.Store.__init__>`
        """
        super(ContextStore, self).__init__(**kwargs)
        self._memory_store = None
        self._include_stored = include_stored
        self._imports_graph = imports_graph
        if context is not None:
            self._init_store(context)

    def open(self, configuration, create=False):
        if self.ctx is not None:
            return VALID_STORE
        else:
            return NO_STORE

    def _init_store(self, ctx):
        self.ctx = ctx

        if self._include_stored:
            self._store_store = RDFContextStore(ctx, imports_graph=self._imports_graph)
        else:
            self._store_store = None

        if self._memory_store is None:
            self._memory_store = IOMemory()
            self._init_store0(ctx)

    def _init_store0(self, ctx, seen=None):
        if seen is None:
            seen = set()
        ctxid = ctx.identifier
        if ctxid in seen:
            return
        seen.add(ctxid)
        self._memory_store.addN((s, p, o, ctxid)
                                for s, p, o
                                in ctx.contents_triples()
                                if not (isinstance(s, Variable) or
                                        isinstance(p, Variable) or
                                        isinstance(o, Variable)))
        for cctx in ctx.imports:
            self._init_store0(cctx, seen)

    def close(self, commit_pending_transaction=False):
        self.ctx = None
        self._memory_store = None

    # RDF APIs
    def add(self, triple, context, quoted=False):
        raise NotImplementedError("This is a query-only store")

    def addN(self, quads):
        raise NotImplementedError("This is a query-only store")

    def remove(self, triple, context=None):
        raise NotImplementedError("This is a query-only store")

    def triples(self, triple_pattern, context=None):
        if self._memory_store is None:
            raise ContextStoreException("Database has not been opened")
        context = getattr(context, 'identifier', context)
        context_triples = []
        if self._store_store is not None:
            context_triples.append(self._store_store.triples(triple_pattern,
                                                             context))
        return chain(self._memory_store.triples(triple_pattern, context),
                     *context_triples)

    def __len__(self, context=None):
        """
        Number of statements in the store. This should only account for non-
        quoted (asserted) statements if the context is not specified,
        otherwise it should return the number of statements in the formula or
        context given.

        :param context: a graph instance to query or None

        """
        if self._memory_store is None:
            raise ContextStoreException("Database has not been opened")
        if self._store_store is None:
            return len(self._memory_store)
        else:
            # We don't know which triples may overlap, so we can't return an accurate count without doing something
            # expensive, so we just give up
            raise NotImplementedError()

    def contexts(self, triple=None):
        """
        Generator over all contexts in the graph. If triple is specified,
        a generator over all contexts the triple is in.

        if store is graph_aware, may also return empty contexts

        :returns: a generator over Nodes
        """
        if self._memory_store is None:
            raise ContextStoreException("Database has not been opened")
        seen = set()
        rest = ()

        if self._store_store is not None:
            rest = self._store_store.contexts(triple)

        for ctx in chain(self._memory_store.contexts(triple), rest):
            if ctx in seen:
                continue
            seen.add(ctx)
            yield ctx


class RDFContextStore(Store):
    # Returns triples imported by the given context
    context_aware = True

    def __init__(self, context=None, imports_graph=None, include_imports=True, **kwargs):
        super(RDFContextStore, self).__init__(**kwargs)
        self.__graph = context.rdf
        self.__imports_graph = imports_graph
        self.__store = self.__graph.store
        self.__context = context
        self.__context_transitive_imports = None
        self.__include_imports = include_imports

    def __init_contexts(self):
        if self.__store is not None and self.__context_transitive_imports is None:
            if not self.__context or self.__context.identifier is None:
                self.__context_transitive_imports = {getattr(x, 'identifier', x)
                                                     for x in self.__store.contexts()}
            elif self.__include_imports:
                imports = transitive_lookup(self.__store,
                                            self.__context.identifier,
                                            CONTEXT_IMPORTS,
                                            self.__imports_graph)
                self.__context_transitive_imports = imports
            else:
                # XXX we should maybe check that the provided context actually exists in
                # the backing graph -- at this point, it's more-or-less assumed in this
                # case though if self.__include_imports is True, we could have an empty
                # set of imports => we query against everything
                self.__context_transitive_imports = set([self.__context.identifier])
            imports_id = getattr(self.__imports_graph, 'identifier', self.__imports_graph)
            if imports_id is not None:
                self.__context_transitive_imports.add(imports_id)

    def triples(self, pattern, context=None):
        self.__init_contexts()

        context = getattr(context, 'identifier', context)
        ctx = None if context is None else self.__graph.get_context(context)
        for t in self.__store.triples(pattern, ctx):
            contexts = set(getattr(c, 'identifier', c) for c in t[1])
            if self.__context_transitive_imports:
                inter = self.__context_transitive_imports & contexts
            else:
                inter = contexts
            if inter:
                yield t[0], inter

    def remove(self, pattern, context=None):
        self.__init_contexts()

        context = getattr(context, 'identifier', context)
        ctx = None if context is None else self.__graph.get_context(context)
        for t in self.__store.triples(pattern, ctx):
            triple = t[0]
            contexts = set(getattr(c, 'identifier', c) for c in t[1])
            if self.__context_transitive_imports:
                inter = self.__context_transitive_imports & contexts
            else:
                inter = contexts
            for ctx in inter:
                self.__store.remove((triple[0], triple[1], triple[2]), ctx)

    def triples_choices(self, pattern, context=None):
        self.__init_contexts()

        ctx = None if context is None else self.__graph.get_context(context)
        for t in self.__store.triples_choices(pattern, ctx):
            contexts = set(getattr(c, 'identifier', c) for c in t[1])
            if self.__context_transitive_imports:
                inter = self.__context_transitive_imports & contexts
            else:
                inter = contexts

            if inter:
                yield t[0], inter

    def contexts(self, triple=None):
        if triple is not None:
            for x in self.triples(triple):
                for c in x[1]:
                    yield getattr(c, 'identifier', c)
        else:
            self.__init_contexts()
            for c in self.__context_transitive_imports:
                yield c

    def namespace(self, prefix):
        return self.__store.namespace(prefix)

    def prefix(self, uri):
        return self.__store.prefix(uri)

    def bind(self, prefix, namespace):
        return self.__store.bind(prefix, namespace)

    def namespaces(self):
        for x in self.__store.namespaces():
            yield x



import functools


@functools.total_ordering
class Quantity:
    _ur = None  # Unit registry

    @classmethod
    def ur(cls):
        # importing pint and initialiing the UnitRegistry adds about a second
        # to the import time, so we delay it until it's needed
        import pint as Q
        if cls._ur is None:
            cls._ur = Q.UnitRegistry()

        return cls._ur

    @classmethod
    def parse(self, s):
        q = self.ur().Quantity(s)
        my_q = Quantity(0, "mL")
        my_q._quant = q
        return my_q

    def __init__(self, value, unit):
        self._quant = self.ur().Quantity(value, unit)

    @property
    def unit(self):
        return str(self._quant.units)

    @property
    def value(self):
        return self._quant.magnitude

    def __str__(self):
        return str(self._quant)

    def __repr__(self):
        return repr(self._quant)

    def __eq__(self, other):
        if isinstance(other, Quantity):
            return (id(self) == id(other)) or (self._quant == other._quant)
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, Quantity):
            return self._quant < other._quant
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self._quant)



import rdflib
from .utils import FCN
from .graph_object import GraphObject, IdentifierMissingException


class Variable(GraphObject):
    '''
    A graph object representing a variable. Typically used in property values
    '''
    def __init__(self, name, **kwargs):
        super(Variable, self).__init__()
        self.var = rdflib.Variable(name)

    @property
    def identifier(self):
        raise IdentifierMissingException(self)

    def variable(self):
        return self.var

    @property
    def defined(self):
        return False

    def __hash__(self):
        return hash(self.var)

    def __str__(self):
        return str(self.var)

    def __repr__(self):
        return '{}({})'.format(FCN(type(self)), repr(self.var))



from os import scandir
from os.path import join as p, exists
import errno

try:
    from urllib.parse import quote as urlquote
except ImportError:
    from urllib import quote as urlquote

from .exceptions import NotABundlePath, BundleNotFound


BUNDLE_MANIFEST_VERSION = 1
'''
Current version number of the bundle manifest. Written by `Installer` and anticipated by
`Deployer` and `Fetcher`.
'''

BUNDLE_ARCHIVE_MIME_TYPE = 'application/x-gtar'
'''
MIME type for bundle archive files
'''


BUNDLE_INDEXED_DB_NAME = 'owm.db'
'''
Base name of the indexed database that gets built in a bundle directory during
installation
'''

BUNDLE_MANIFEST_FILE_NAME = 'manifest'
'''
Name of the manifest file in a bundle directory or archive
'''


def fmt_bundle_directory(bundles_directory, ident, version=None):
    '''
    Get the directory for the given bundle identifier and version

    Parameters
    ----------
    ident : str
        Bundle identifier
    version : int
        Version number. If not provided, returns the directory containing all of the
        versions
    '''
    base = p(bundles_directory, urlquote(ident, safe=''))
    if version is not None:
        return p(base, str(version))
    else:
        return base


def validate_manifest(bundle_path, manifest_data):
    '''
    Validate manifest data in a `dict`

    Parameters
    ----------
    bundle_path : str
        The path to the bundle directory or archive. Used in the exception message if the
        manifest data is invalid
    manifest_data : dict
        The data from a manifest file

    Raises
    ------
    NotABundlePath
        Thrown in one of these conditions:

        - `manifest_data` lacks a `manifest_version`
        - `manifest_data` has a `manifest_version` > BUNDLE_MANIFEST_VERSION
        - `manifest_data` has a `manifest_version` <= 0
        - `manifest_data` lacks a `version`
        - `manifest_data` lacks an `id`
    '''
    manifest_version = manifest_data.get('manifest_version')
    if not manifest_version:
        raise NotABundlePath(bundle_path,
                'the bundle manifest has no manifest version')

    if manifest_version > BUNDLE_MANIFEST_VERSION or manifest_version <= 0:
        raise NotABundlePath(bundle_path,
                'the bundle manifest has an invalid manifest version')

    version = manifest_data.get('version')
    if not version:
        raise NotABundlePath(bundle_path,
                'the bundle manifest has no bundle version')

    ident = manifest_data.get('id')
    if not ident:
        raise NotABundlePath(bundle_path,
                'the bundle manifest has no bundle id')


def find_bundle_directory(bundles_directory, ident, version=None):
    # - look up the bundle in the bundle cache
    # - generate a config based on the current config load the config
    # - make a database from the graphs, if necessary (similar to `owm regendb`). If
    #   delete the existing database if it doesn't match the store config
    if version is None:
        bundle_root = fmt_bundle_directory(bundles_directory, ident)
        latest_version = 0
        try:
            ents = scandir(bundle_root)
        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT: # FileNotFound
                raise BundleNotFound(ident, 'Bundle directory does not exist') from e
            raise

        for ent in ents:
            if ent.is_dir():
                try:
                    vn = int(ent.name)
                except ValueError:
                    # We may put things other than versioned bundle directories in
                    # this directory later, in which case this is OK
                    pass
                else:
                    if vn > latest_version:
                        latest_version = vn
        version = latest_version
    if not version:
        raise BundleNotFound(ident, 'No versioned bundle directories exist')
    res = fmt_bundle_directory(bundles_directory, ident, version)
    if not exists(res):
        if version is None:
            raise BundleNotFound(ident,
                    f'Bundle directory, "{res}", does not exist')
        else:
            raise BundleNotFound(ident,
                    f'Bundle directory, "{res}", does not exist for the specified version', version)
    return res
